from dataclasses import dataclass, field
import importlib

from rich.console import Console
console = Console()

DEBUG_STYLES = False

@dataclass
class SSH_Host:
    """ Class for SSH host config structure """
    name: str
    group: str
    type: str = "normal"
    info: list = field(default_factory=list)
    params: dict = field(default_factory=dict)

    inherited_params: list = field(default_factory=list)
    print_style: str = "panels"


    # Method for interaction with printing the object via Rich library
    # Each supported style should be located in defined folder (by default under "host_styles")
    # Each module must have "render" function, and return "rich renderable" object
    def __rich__(self):
        try:
            # If current host "print_style" for is set to "panels", we will try
            # to import module from "./host_styles/panels.py"
            style = importlib.import_module(f'sshclick.sshc.host_styles.{self.print_style}')
            return style.render(self)
        except ModuleNotFoundError:
            return f"SSH_Host style [bright_red]'{self.print_style}'[/] is not implemented!"
        except Exception:
            if DEBUG_STYLES:
                console.print_exception(show_locals=False, max_frames=1, suppress=[importlib])
            return f"SSH_Host style [bright_red]'{self.print_style}'[/] is broken!"



import click
from sshclick.sshc import SSH_Config, complete_ssh_host_names

#------------------------------------------------------------------------------
# COMMAND: host delete
#------------------------------------------------------------------------------
@click.command(name="delete", help="Delete host from configuration")
@click.argument("name", shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx, name):
    config: SSH_Config = ctx.obj

    found_host, found_group = config.find_host_by_name(name, throw_on_fail=False)
    if not found_host:
        print(f"Cannot delete host '{name}' as it is not defined in configuration!")
        ctx.exit(1)
    
    if found_host.type == "normal":
        found_group.hosts.remove(found_host)
    else:
        found_group.patterns.remove(found_host)

    config.generate_ssh_config().write_out()
    if not config.stdout:
        print(f"Deleted host: {name}")



from dataclasses import dataclass, field

@dataclass
class SSH_Group:
    """ Class for SSH Group config structure """
    name: str
    desc: str = ""
    info: list = field(default_factory=list)
    hosts: list = field(default_factory=list)
    patterns: list = field(default_factory=list)

    print_style: str = ""

    def __rich__():
        pass





# from rich.table import Table
from prettytable import PrettyTable

#TODO: Try to rework with Rich components !!!
def generate_graph(traced_hosts: list):
    """
    Function that generates nice "graph" view of connected hosts
    """

    # TODO: This is currently static, but we could improve it, as in very long hostname
    #       it will result in broken lines, or alternatively in cut-off names
    max_cell_size = 24   # Keep value as multiple of 2!

    def crop_str(string):
        string = (string[:max_cell_size-2] + '..') if len(string) > max_cell_size else string
        return string

    reversed_hosts = traced_hosts[::-1]
    num_elements = len(traced_hosts) + 1

    # We use prettytable as way to organize graph in columns then print it
    # without usual decorators such as borders and headers, to get graph
    trace = PrettyTable()
    trace.add_column("local", [
        " Connection graph   ",
        "  ────────────────  ",
        "         THIS HOST  ",
        "    SSH Connection ─",
        "                    ",
    ], align="r")

    last_index = len(traced_hosts) - 1
    cp = int((max_cell_size - 6) / 2)  #cell_padding
    for i, host in enumerate(reversed_hosts):
        if i != last_index:
            # Add Jump proxies
            trace.add_column(host.name, [
                crop_str(host.name),                       # SSH Host     (could be long name)
                crop_str(host.params.get("hostname", "")),                   # SSH HostName (usually IP, but can also be long fqdn)
                " " * cp + "│    │" + " " * cp,
                "─" * cp + "╯    ╰" + "─" * cp,
                "",
            ])
        else:
            # Target (last) element
            trace.add_column("target", [
                "         ",
                "         ",
                "  TARGET",
                "─ " + host.name,
                "  " + host.params.get("hostname", ""),
            ], align="l")

    # Print local forward tunnel between current host and target host
    if "localforward" in traced_hosts[0].params:
        # trace.add_row(["" for _ in range(num_elements)])   # Empty row
        trace.add_row(["Local SSH tunnels  "] + ["" for _ in range(num_elements - 1)])
        
        for tun in traced_hosts[0].params["localforward"]:
            tunnel_row = []
            tun_port, tun_end = tun.split()
            for i in range(num_elements):
                if i == 0:                                      # Fill source column (aligned R)
                    tunnel_row.append(f":{tun_port} >")
                elif i != num_elements - 1:                     # Add Jump proxies   (aligned M)
                    tunnel_row.append(">" * max_cell_size)
                else:                                           # Fill last/target   (aligned L)
                    tunnel_row.append("> " + tun_end)
            trace.add_row(tunnel_row)

    # Print remote forward tunnel between current host and target host
    if "remoteforward" in traced_hosts[0].params:
        # trace.add_row(["" for _ in range(num_elements)])   # Empty row
        trace.add_row(["Remote SSH tunnels:  "] + ["" for _ in range(num_elements - 1)])
        
        for tun in traced_hosts[0].params["remoteforward"]:
            tunnel_row = []
            tun_port, tun_end = tun.split()
            for i in range(num_elements):
                if i == 0:                                      # Fill source column (aligned R)
                    tunnel_row.append(f"{tun_end} <")
                elif i != num_elements - 1:                     # Add Jump proxies   (aligned M)
                    tunnel_row.append("<" * max_cell_size)
                else:                                           # Fill last/target   (aligned L)
                    tunnel_row.append("< :" + tun_port)
            trace.add_row(tunnel_row)

    # Remove all decorations from cells
    trace.header = False
    trace.border = False
    trace.padding_width = 0
    return trace.get_string()




import click
from sshclick.sshc import SSH_Config, complete_ssh_group_names

#TODO: Check click.edit for multiline edit option (info, or even params?)

#------------------------------------------------------------------------------
# COMMAND: group-set
#------------------------------------------------------------------------------
@click.command(name="set", help="Change group parameters")
@click.option("-r", "--rename", default=None, help="Rename group")
@click.option("-d", "--desc", default=None, help="Set description")
@click.option("-i", "--info", default=None, multiple=True, help="Set info, can be set multiple times")
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.pass_context
def cmd(ctx, name, rename, desc, info):
    config: SSH_Config = ctx.obj

    # Nothing was provided
    if not rename and not desc and not info:
        print("Calling set on group without specifying any option is not valid!")
        print("Run 'sshc group set -h' for help.")
        ctx.exit(1)

    # Find group by name
    found_group = config.find_group_by_name(name, throw_on_fail=False)
    if not found_group:
        print(f"Cannot modify group '{name}', it is not defined in configuration!")
        ctx.exit(1)

    if rename:
        found_group.name = rename
    
    if desc:
        found_group.desc = desc

    if info:
        if len(info[0]) > 0:
            found_group.info = info
        else:
            found_group.info = []

    config.generate_ssh_config().write_out()
    if not config.stdout:
        print(f"Modified group: {name}")



import click
import re
import copy
from sshclick.sshc import SSH_Config, SSH_Group, SSH_Host

from rich.console import Console
from rich.table import Table
from rich import box

console = Console()

#------------------------------------------------------------------------------
# COMMAND: host-list
#------------------------------------------------------------------------------
@click.command(name="list", help="List hosts")
@click.option("-g", "--group",  "group_filter", default=None,  help="Filter for group that host belongs to (regex)")
@click.option("-n", "--name",   "name_filter",  default=None,  help="Filter for host name (regex)")
@click.option("-v", "--verbose", is_flag=True,  default=False, help="Show verbose info (all parameters)")
@click.pass_context
def cmd(ctx, group_filter, name_filter, verbose):
    config: SSH_Config = ctx.obj

    # Filter out groups and hosts if filters are defined via CLI
    filtered_groups: list[SSH_Group] = []
    for group in config.groups:
        # If group filter is defined, check if current group matches the name to progress
        if group_filter:
            group_match = re.search(group_filter, group.name)
            if not group_match:
                continue
        
        # When group is not skipped, check if name filter is used, and filter out groups
        if name_filter:
            # Make a new copy of group, so we dont mess original config
            group_copy = copy.copy(group)
            group_copy.hosts = []
            group_copy.patterns = []
            include_group = False

            for host in group.hosts + group.patterns:
                match = re.search(name_filter, host.name)
                if match:
                    include_group = True
                    if host.type == "normal":
                        group_copy.hosts.append(host)
                    else:
                        group_copy.patterns.append(host)
            if include_group:
                filtered_groups.append(group_copy)
        else:
            filtered_groups.append(group)

    if not filtered_groups:
        print("No host is matching any given filter!")
        ctx.exit(1)
    
    # This lists define host properties and which parameters will be displayed
    host_props = ["name", "group", "type"]
    params     = ["hostname", "user"]

    # If output is verbose, we need to find all parameters, and add them to params list
    if verbose:
        flat_config: list[SSH_Host] = []
        for group in filtered_groups:
            for h in group.hosts + group.patterns:
                flat_config.append(h)
        for host in flat_config:
            for i_params in host.params:
                if (not i_params in params):
                    params.append(i_params)
    
    header = host_props + ([f"param:{p}" for p in params])
    table = Table(*header, box=box.SQUARE, style="gray35")

    # Start adding rows    
    for group in filtered_groups:
        # Iterate trough hosts and patters
        for host in group.hosts + group.patterns:
            inherited: list[tuple[str, dict]] = []
            if host.type == "normal":
                inherited = config.find_inherited_params(host.name)
            host_params = []
            # Go trough list of all params we know are available across current host list
            for table_param in params:
                if table_param in host.params:
                    # Handle direct params, and handle if its a list or string
                    if isinstance(host.params[table_param], list):
                        host_params.append("\n".join(host.params[table_param]))
                    else:
                        host_params.append(host.params[table_param])
                else:
                    # Handle inherited params (only valid for "normal" hosts)
                    if inherited:
                        for pattern, i_params in inherited:
                            if table_param in i_params:
                                if isinstance(i_params[table_param], list):
                                    table_param = "\n".join([f"{val}  ({pattern})" for val in i_params[table_param]])
                                    host_params.append(table_param)
                                else:
                                    host_params.append(f"[yellow]{i_params[table_param]}  ({pattern})[/]")
                            else:
                                host_params.append("")
                    else:
                        host_params.append("")
            row = [host.__dict__[prop] for prop in host_props] + host_params
            table.add_row(*row) if host.type == "normal" else table.add_row(*row, style="cyan")

    console.print(table)



import click
from .globals import *
from .sshc import SSH_Config

# Setup click to use both short and long help option
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--sshconfig", default=DEFAULT_USER_CONF, help="Config file, default is ~/.ssh/config")
@click.option("--stdout",    default=DEFAULT_STDOUT, is_flag=True, help="Send changed SSH config to STDOUT instead to original file")
@click.version_option(VERSION, message="SSHClick (sshc) - Version: %(version)s")
@click.pass_context
def cli(ctx, sshconfig: str, stdout: bool):
    """
    SSHClick - SSH Config manager

    Note: As this is early alpha, backup your SSH config files before
    this software, as you might accidentally lose some configuration
    """
    # Prepare Click context object
    ctx.obj = SSH_Config(file=sshconfig, stdout=stdout).read().parse()


# Top full commands
from .cmds import cmd_group, cmd_host
cli.add_command(cmd_host.ssh_host)
cli.add_command(cmd_group.ssh_group)



import click
from sshclick.sshc import SSH_Config
from rich.console import Console
from rich.table import Table
from rich import box

#------------------------------------------------------------------------------
# COMMAND: group list
#------------------------------------------------------------------------------
@click.command(name="list", help="Lists all groups")
@click.pass_context
def cmd(ctx):
    config: SSH_Config = ctx.obj

    table = Table(box=box.SQUARE, style="grey35")
    table.add_column("Name", style="white")
    table.add_column("Num.Hosts", justify="right", style="bright_yellow")
    table.add_column("Desc", style="gray50")

    for group in config.groups:
        table.add_row(group.name, str(len(group.hosts)), group.desc)

    console = Console()
    console.print(table)



import click
from sshclick.globals import *
from sshclick.sshc import SSH_Config, SSH_Group, SSH_Host

#------------------------------------------------------------------------------
# COMMAND: host create
#------------------------------------------------------------------------------
@click.command(name="create", help="Create new host configuration")
@click.option("-g", "--group", "target_group_name", default=DEFAULT_GROUP_NAME, help="Sets group for the host")
@click.option("-p", "--parameter", default=[], multiple=True, help="Sets parameter for the host, must be in 'param=value' format")
@click.option("--force", is_flag=True, default=False, help="Forces automatically create group for host if group is missing")
@click.argument("name")
@click.pass_context
def cmd(ctx, name, target_group_name, parameter, force):
    config: SSH_Config = ctx.obj

    found_host, _ = config.find_host_by_name(name, throw_on_fail=False)
    if found_host:
        print(f"Cannot create host '{name}' as it already exists in configuration!")
        ctx.exit(1)
    
    # Find group by name where to store config
    found_group = config.find_group_by_name(target_group_name, throw_on_fail=False)
    if not found_group:
        if not force:
            print(f"Cannot create host '{name}' in group '{target_group_name}' since the group does not exist")
            print("Create group first, or use '--force' option to create it automatically!")
            ctx.exit(1)
        else:
            # Create new group
            new_group = SSH_Group(name=target_group_name)
            # Add new group to config and set it as target group
            config.groups.append(new_group)
            found_group = new_group

    # This is patter host
    target_type = "pattern" if "*" in name else "normal"
    new_host = SSH_Host(name=name, group=target_group_name, type=target_type)

    # Add all passed parameters to config
    # TODO: here we need to implement validation for all "known" parameters that can be used, and possibly
    # some normalization to documented "CamelCase" format
    for item in parameter:
        param, value = item.split("=")
        param = param.lower()           # parametar keyword will be lowercased as they are case insensitive
        if param in PARAMS_WITH_ALLOWED_MULTIPLE_VALUES:
            # We need to handle host parameter as "list"
            if not param in new_host.params:
                new_host.params[param] = [ value ]
            else:
                new_host.params[param].append(value)
        else:
            # Simple single keyword
            new_host.params[param] = value

    if new_host.type == "normal":
        found_group.hosts.append(new_host)
    else:
        found_group.patterns.append(new_host)

    config.generate_ssh_config().write_out()
    if not config.stdout:
        print(f"Created host: {name}")



import click
from sshclick.sshc import SSH_Config, complete_ssh_group_names

from rich.table import Table
from rich.panel import Panel
from rich import box
from rich.console import Console

#------------------------------------------------------------------------------
# COMMAND: group show
#------------------------------------------------------------------------------
@click.command(name="show", help="Shows group details")
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.pass_context
def cmd(ctx, name):
    config: SSH_Config = ctx.obj

    found_group = config.find_group_by_name(name, throw_on_fail=False)
    if not found_group:
        print(f"Cannot show group '{name}', it is not defined in configuration!")
        ctx.exit(1)

    host_list = []
    for host in found_group.hosts:
        if "hostname" in host.params:
            h_name = host.params["hostname"]
            h_port = (":" + host.params["port"]) if "port" in host.params else ""
            host_list.append(f"{host.name} ({h_name}{h_port})")
        else:
            host_list.append(host.name)

    pattern_list = []
    for host in found_group.patterns:
        # hack to search via case insensitive info
        if "hostname" in host.params:
            h_name = host.params["hostname"]
            pattern_list.append(f"{host.name} ({h_name})")
        else:
            pattern_list.append(host.name)


    # New rich formatting
    table = Table(box=box.SQUARE, style="grey35")
    table.add_column("Group Parameter", no_wrap=True)
    table.add_column("Value")

    table.add_row("name",        found_group.name,                   style="white")
    table.add_row("description", found_group.desc,                   style="grey50")
    table.add_row("info",        Panel("\n".join(found_group.info)), style="grey50")
    table.add_row("hosts",       Panel("\n".join(host_list)),         style="white")
    table.add_row("patterns",    Panel("\n".join(pattern_list)),      style="cyan")

    console = Console()
    console.print(table)




from ..ssh_host import SSH_Host

from rich.table import Table
from rich.panel import Panel
from rich import box

#------------------------------------------------------------------------------
# Render host data in panel with single-card look
#------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"
    out_info = "\n".join(host.info) if host.info else "- No info defined - "

    layout_table = Table("", box=box.ROUNDED, style="grey35" ,show_edge=True, show_header=False)
    layout_table.add_row(f"[bright_white]Name [/]:  {host.name}")
    layout_table.add_row(f"[bright_white]Group[/]:  {host.group}")
    layout_table.add_row(f"[bright_white]Type [/]:  {out_type}")

    if host.info:
        layout_table.add_row("")
        layout_table.add_row(f"[gray50]{out_info}[/]")

    layout_table.add_row("")

    param_table = Table(box=box.SIMPLE, style="grey35", show_header=True, show_edge=False, pad_edge=False)
    param_table.add_column("Param")
    param_table.add_column("Value")
    param_table.add_column("Inherited-from")

    # Add rows for SSH Config parameters
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        param_table.add_row(key, output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if not param in host.params:
                output_value = value if not isinstance(value, list) else "\n".join(value)
                param_table.add_row(param, output_value, pattern, style="yellow")

    layout_table.add_row(param_table)
    wrapper_panel = Panel(layout_table, box=box.SIMPLE, padding=(0,0))

    return wrapper_panel



import click

#------------------------------------------------------------------------------
# GROUP Commands
#------------------------------------------------------------------------------
@click.group(name="group")
def ssh_group():
    """
    Manage groups
    """

#// Linking other sub-commands
from .group import group_create, group_delete, group_list, group_set, group_show
ssh_group.add_command(group_create.cmd)
ssh_group.add_command(group_delete.cmd)
ssh_group.add_command(group_list.cmd)
ssh_group.add_command(group_set.cmd)
ssh_group.add_command(group_show.cmd)



from .ssh_config import SSH_Config
from .ssh_group import SSH_Group
from .ssh_host import SSH_Host

from .sshextras import generate_graph
from .sshutils import complete_ssh_host_names
from .sshutils import complete_ssh_group_names



import os.path

VERSION = "0.4.4"

# Setup defaults
DEFAULT_USER_CONF = os.path.expanduser("~/.ssh/config")
DEFAULT_GROUP_NAME = "default"
DEFAULT_STDOUT = False

# Update this as needed
PARAMS_WITH_ALLOWED_MULTIPLE_VALUES = [
    "localforward",
    "remoteforward",
    "dynamicforward",
]



import click

#------------------------------------------------------------------------------
# HOST Commands
#------------------------------------------------------------------------------
@click.group(name="host")
def ssh_host():
    """
    Manage hosts
    """

#// Linking other sub-commands
from .host import host_create, host_delete, host_list, host_set, host_show, host_test
ssh_host.add_command(host_create.cmd)
ssh_host.add_command(host_delete.cmd)
ssh_host.add_command(host_list.cmd)
ssh_host.add_command(host_set.cmd)
ssh_host.add_command(host_show.cmd)
ssh_host.add_command(host_test.cmd)



import click
import subprocess
from sshclick.sshc import SSH_Config, complete_ssh_host_names

#------------------------------------------------------------------------------
# COMMAND: host connect
#------------------------------------------------------------------------------
@click.command(name="test", help="Test SSH host connection")
@click.option("--timeout", default=3, help="Timeout for SSH connection")
@click.argument("name", shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx, name, timeout):
    config: SSH_Config = ctx.obj

    found_host, _ = config.find_host_by_name(name, throw_on_fail=False)
    if not found_host:
        print(f"Cannot test host '{name}' as it is not defined in configuration!")
        ctx.exit(1)

    hostname = found_host.params["hostname"]

    #// SSH Command method via subprocess
    #------------------------------------------------
    ssh_command = f"ssh -q -o StrictHostKeyChecking=no -o ConnectTimeout={timeout} {name} 'exit 0'"
    response = subprocess.run(ssh_command, shell=True)
    
    print(f"Connection with SSH Host name: {name} ({hostname}) ", end="")
    if response.returncode == 0:
        print("successful!")
        ctx.exit()
    else:
        print("failed!")
        ctx.exit(1)



import click
from sshclick.sshc import SSH_Config, complete_ssh_group_names

#------------------------------------------------------------------------------
# COMMAND: group delete
#------------------------------------------------------------------------------
@click.command(name="delete", help="Delete group")
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.pass_context
def cmd(ctx, name):
    config: SSH_Config = ctx.obj

    # Find group by name
    found_group = config.find_group_by_name(name, throw_on_fail=False)
    if not found_group:
        print(f"Cannot delete group '{name}', it is not defined in configuration!")
        ctx.exit(1)
    
    config.groups.remove(found_group)

    config.generate_ssh_config().write_out()
    if not config.stdout:
        print(f"Deleted group: {name}")



import os.path
from .ssh_config import SSH_Config
from rich.console import Console
err = Console(stderr=True)

# Make a copy of input dict with all keys as LC and filtered out based on input filter list
def filter_dict(d: dict, ignored: list = []) -> dict:
    return {k: v for (k, v) in d.items() if k not in ignored}


# Custom parsing trough parent object types until required parameters are found
# Then build config object and bound it to ctx.obj
def build_context_config(ctx):
    if ctx.obj is None:
        current_obj = ctx.parent
        try:
            while True:
                if "sshconfig" in current_obj.params:
                    full_path = os.path.expanduser(current_obj.params["sshconfig"])
                    ctx.obj = SSH_Config(
                        file=full_path,
                        stdout=current_obj.params["stdout"]
                    ).read().parse()
                    break
                else:
                    current_obj = current_obj.parent
        except:
            err.print("INTERNAL ERROR: Could not reconstruct context for SSH configuration!")
            ctx.exit(1)


# For some reason I cant get context object initialized by main app when running autocomplete
# BUG: https://github.com/pallets/click/issues/2303
def complete_ssh_host_names(ctx, param, incomplete):
    build_context_config(ctx)

    all_hosts = ctx.obj.get_all_host_names()
    return [k for k in all_hosts if k.startswith(incomplete)]


# For some reason I cant get context object initialized by main app when running autocomplete
# BUG: https://github.com/pallets/click/issues/2303
def complete_ssh_group_names(ctx, param, incomplete):
    build_context_config(ctx)

    all_groups = ctx.obj.get_all_group_names()
    return [k for k in all_groups if k.startswith(incomplete)]





import click
from sshclick.sshc import SSH_Config, SSH_Group

#------------------------------------------------------------------------------
# COMMAND: group create
#------------------------------------------------------------------------------
@click.command(name="create", help="Create new group")
@click.option("-d", "--desc", default=None, type=str, help="Short description of group")
@click.option("-i", "--info", default=[], multiple=True, type=str, help="Info line, can be set multiple times")
@click.argument("name")
@click.pass_context
def cmd(ctx, name, desc, info):
    config: SSH_Config = ctx.obj

    # Check if already group exists
    found_group = config.find_group_by_name(name, throw_on_fail=False)
    if found_group:
        print(f"Cannot create new group '{name}', as group already exists with this name")
        ctx.exit(1)

    new_group = SSH_Group(name, desc=desc, info=info)

    # Add new group to config and show newly created group
    config.groups.append(new_group)
    
    config.generate_ssh_config().write_out()
    if not config.stdout:
        print(f"Created group: {name}")
    


import click
from sshclick.sshc import SSH_Config, complete_ssh_host_names, generate_graph

from rich.console import Console
console = Console()

# Help autocomplete with host styles
enabled_styles = ["panels", "card", "simple", "table", "table2", "json"]
def complete_styles(ctx, param, incomplete):
    return [k for k in enabled_styles if k.startswith(incomplete)]

#------------------------------------------------------------------------------
# COMMAND: host show
#------------------------------------------------------------------------------
@click.command(name="show", help="Display info for host")
@click.option("--follow", is_flag=True, help="Follow and displays all connected jump-proxy hosts")
@click.option("--graph",  is_flag=True, help="Shows connection to target as graph (default:false)")
@click.option("--style", default="panels", show_default="panels", envvar='SSHC_HOST_STYLE',
    help="Select output rendering style for host details: [panels, card, simple, table, table2, json]",
    shell_complete=complete_styles)
@click.argument("name", shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx, name, follow, graph, style):
    config: SSH_Config = ctx.obj

    # Keep list of linked hosts via jumpproxy option
    traced_hosts = []

    # we are first checking host that is asked, and then check if that host has a "proxy" defined
    # if proxy is defined, we add found host to traced list, and search attached proxy, it this way we
    # can trace full connection path for later graph processing
    while True:
        # Search for host in current configuration
        found_host, _ = config.find_host_by_name(name,throw_on_fail=False)
        if not found_host:
            print(f"Cannot show host '{name}' as it is not defined in configuration!")
            ctx.exit(1)

        #TODO: Should this be part of configuration parsing stage?
        inherited_params: list[tuple[str, dict]] = []
        if found_host.type == "normal":
            inherited_params = config.find_inherited_params(name)
            found_host.inherited_params = inherited_params

        # Set SSH host print style from user input
        found_host.print_style = style

        # Add current host, and its table to traced lists
        traced_hosts.append(found_host)

        # Find proxy info if it exists, if not, break the loop
        proxyjump = None
        if "proxyjump" in found_host.params:
            proxyjump = found_host.params["proxyjump"]
        else:
            for _, params in inherited_params:
                for key, value in params.items():
                    if key == "proxyjump":
                        proxyjump = value

        # If there is connected host, we switch "name" and continue the loop
        # Otherwise we exit the loop
        if proxyjump:
            name = proxyjump
        else:
            break
    
    # Print collected host data
    if not follow:
        console.print(traced_hosts[0])
    else:
        for i, output in enumerate(traced_hosts):
            console.print(output)
            if i < len(traced_hosts) - 1:
                console.print("  [bright_green]▼ via proxy[/]")
    
    #TODO: Make better graph output
    if graph:
        console.print("\n", generate_graph(traced_hosts), "\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sshclick',
 'sshclick.cmds',
 'sshclick.cmds.group',
 'sshclick.cmds.host',
 'sshclick.sshc',
 'sshclick.sshc.host_styles']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0', 'prettytable>=3.2.0,<4.0.0', 'rich>=12.5.1,<13.0.0']

entry_points = \
{'console_scripts': ['sshc = sshclick.main:cli']}

setup_kwargs = {
    'name': 'sshclick',
    'version': '0.4.4',
    'description': 'SSH Config manager',
    'long_description': '# SSH Click Config manager (sshclick)\n\nCheck this page on [GitHub](https://github.com/karlot/sshclick)\n\n## Links\n\n- [Intro](https://github.com/karlot/sshclick#intro)\n- [Why?](https://github.com/karlot/sshclick#why)\n- [What does it do](https://github.com/karlot/sshclick#what-does-it-do)\n  - [Installation procedure](https://github.com/karlot/sshclick#installation-procedure)\n  - [Upgrade procedure](https://github.com/karlot/sshclick#upgrade-procedure)\n  - [Uninstall procedure](https://github.com/karlot/sshclick#uninstall-procedure)\n- [SSH Config structure](https://github.com/karlot/sshclick#ssh-config-structure-and-important-note-about-comments)\n  - [Comment blocks and metadata](https://github.com/karlot/sshclick#comment-blocks-and-metadata-in-ssh-config)\n- [Example usage and features](https://github.com/karlot/sshclick#example-usage-and-features)\n  - [Group commands and options](https://github.com/karlot/sshclick#group-commands-and-options)\n  - [Host commands and options](https://github.com/karlot/sshclick#host-commands-and-options)\n  - [Output styling and user ENV variables](https://github.com/karlot/sshclick#output-styling-and-user-env-variables)\n- [Recorded demos](https://github.com/karlot/sshclick#recorded-demos)\n  - [sshc group operations](https://github.com/karlot/sshclick#demo-showing-some-sshc-group-operations)\n  - [sshc host operations](https://github.com/karlot/sshclick#demo-showing-some-sshc-host-operations)\n- [Author](https://github.com/karlot/sshclick#author)\n- [License](https://github.com/karlot/sshclick#license)\n\n\n## Intro\nTerminal based assisted management of your SSH config files.  \nBuilt out of boredom with managing messy and huge ssh_config files.  \n\nEARLY VERSION, backup your SSH config files before using!  \nSSHClick can be used with "show" and "list" commands for hosts, without modifying your SSH Config!  \n\n**Only commands that modify configuration will edit and rewrite/restructure your SSH Config file. In that case, any added comment or infos that are not in form that SSHClick understand will be discarded, and configuration will be re-formatted to match SSHClick style. See below details to understand how SSH Click would keep your config organized**\n\n## Why?\n\nWhat am I trying to solve with this tool?\n* I need something that works fast and great in terminal, and does not require complex setup.\n* Managing some other configuration files that renders to SSH config is extra step that I don\'t like.\n* SSH config is already feature-full with all options SSH client support, why inventing extra layer?\n* SSH config is the only config I need to backup.\n* I need quick way to search, group and visualize all hosts inside SSH configuration (especially since it can grow huge)\n\n\n## What does it do?\n\nSSHClick (sshc) is just a tool designed to work with existing SSH configuration files on your Linux/Windows/WSL terminal environment.  \nIt basically parses your SSH config, and can provide easy commands to list, filter, modify or view specific Host entries.\nTrough additional "magic" comments it can add abstractions such as "groups" and various information that is both readable in the configuration file, and can be parsed and printed while using the tool.\n\n\n### Installation procedure\n\nShould be straight forward...  \n\n1. **Check preconditions:**\n    - Currently only tested on Linux (Debian 10,11, Ubuntu 20.04,22.04), but should work on other systems as well\n    - Minimum python3.7 (tested up to 3.11 beta) & pip installed\n        - it is preferable to not use system python version, to install "custom" user python on linux, you can try using pyenv (https://github.com/pyenv/pyenv\n    - git installed\n\n2. **Install package:**\n    - from PyPI using pip\n        ```\n        pip install sshclick\n        ```\n\n    - (OR) from source using pip\n        ```sh\n        git clone https://github.com/karlot/sshclick\n        cd sshclick\n        pip install .\n        ```\n\n3. Use it as you like, `sshc` command should be available to access SSHClick application, see below chapter for basic [usage](https://github.com/karlot/sshclick#example-usage-and-features)\n\n4. Install shell autocompletion (_TAB-TAB auto-completes on commands, options, groups and hosts_)\n    * __Bash__ - Add this line to end of your `~/.profile` file:\n      ```sh\n      eval "$(_SSHC_COMPLETE=bash_source sshc)"\n      ```\n    * __Zsh__ - Add this line to end of your `~/.zshrc` file:\n      ```sh\n      eval "$(_SSHC_COMPLETE=zsh_source sshc)"\n      ```\n\n\n### Upgrade procedure\n\n* Upgrade from new PyPI release:  \n  ```sh\n  pip install --upgrade sshclick\n  ```\n\n* Upgrade from source:\n  Assuming installation is already done, and previous version is cloned in some local folder\n\n  ```sh\n  cd sshclick     # cd into existing previously cloned repo folder\n  git pull\n  pip install --upgrade .\n  ```\n\n\n### Uninstall procedure\nSimply run:\n```\npip uninstall sshclick\n```\n\nIn case you have installed from cloned source code, you can delete locally cloned repo.\n```sh\nrm -r sshclick\n```\n\n---\n\n## SSH Config structure, and important note about comments\n\nSSHClick when editing and writing to SSH config file must use specific style, and is internally using comments to "organize" configuration itself. This means comments outside of what sshclick is handling are unsupported and will be lost when SSHClick modifies a file.)\n\n\n### Comment blocks and metadata in SSH Config\n\nSSHClick uses comments to add extra information which it can use to add concept of grouping and extra information to hosts. Special "metadata" lines start with `#@` followed by some of meta-tags like `group`, `desc`, `info`. This are all considered group metadata tags, as they apply on the group level. Note that line separations above and below "group header" are added only for visual aid, they are ignored at parsing, but are included when modifying/generating SSH config file.  \n\nThis "headers" can be added manually also in SSH config, or sshclick can add them and move hosts under specific group, using `sshc` cli tool\n\nNormally start of the "GROUP HEADER" inside SSH Config would look like below.  \n- `#@group:` is KEY metadata tag, that during "parsing" defines that all hosts configured below this "tag" belong to this group\n- `#@desc:` is optional tag that adds "description" to defined group, and will display in usual group display commands\n- `#@info:` is optional tag that can appear multiple times, adding extra information lines tied to the group.\n\nAdditionally each "host" definition can have optional meta info:\n- `#@host:` is optional tag that can appear multiple times, that can hold some information about the host, this meta info when defined applies to next "host" definition that will appear. If this key is added after "host" keyword, it will be applied to next host, for that reason, keep this host meta info above the actual host definition.\n\nFollowing is sample how group header is rendered by SSHClick:\n```\n#-------------------------------------------------------------------------------\n#@group: <GROUP-NAME>       [MANDATORY]   <-- This line starts new group\n#@desc: <GROUP-DESCRIPTION> [OPTIONAL, SINGLE]\n#@info: <GROUP-INFO-LINES>  [OPTIONAL,MULTIPLE]\n#-------------------------------------------------------------------------------\nHost ...    <-- this hosts definitions are part of the defined group\n    param1 value1\n    param2 value2\n\n#@host: <HOST-INFO-LINES>  [OPTIONAL,MULTIPLE] <-- Adds info to following host\nHost ...\n\n<ANOTHER GROUP HEADER>\n```\n\nIf there are no groups defined, then all hosts are considered to be part of "default" group. SSHClick can be used to move hosts between groups and handle keeping SSH config "tidy" and with consistent format.\n\n\n## Example usage and features\n\nSSHClick is deploying `sshc` cli tool that allows interacting with your SSH Config file and perform various organization,listing, displaying and modification of SSH Group/Host configuration parameters.  \n`sshc` comes with pre-built lots of help options so each level of commands provide `--help` options to provide you info what commands and options are available at which command level.  \n\nFor example to check version, type: `sshc --version`  \n_Sample output:_\n```console\n$ sshc --version\nSSHClick (sshc) - Version: 0.4.1\n```\n\nIf you run `sshc` command alone, or adding `-h` or `--help` option, it will show help what else must be added to the command...  \n_Example:_\n```console\n$ sshc --help\nUsage: sshc [OPTIONS] COMMAND [ARGS]...\n\n  SSHClick - SSH Config manager\n\n  Note: As this is early alpha, backup your SSH config files before this\n  software, as you might accidentally lose some configuration\n\nOptions:\n  --sshconfig PATH  Config file, default is ~/.ssh/config\n  --stdout          Send changed SSH config to STDOUT instead to original file\n  --version         Show current version\n  -h, --help        Show this message and exit.\n\nCommands:\n  group  Manage groups\n  host   Manage hosts\n```\n\nMain sub-commands are for separate control of:  \n- Groups (via `sshc group` command)\n  - Allows operation of groups inside SSH config\n- Hosts (via `sshc host` command)\n  - Allows operation of hosts inside SSH config\n\n### `group` commands and options\n\nTo manage "groups" type `sshc group --help` to see options.  \n_example:_\n```console\n$ sshc group --help\nUsage: sshc group [OPTIONS] COMMAND [ARGS]...\n\n  Manage groups\n\nOptions:\n  -h, --help  Show this message and exit.\n\nCommands:\n  create  Create new group\n  delete  Delete group\n  list    Lists all groups\n  set     Change group parameters\n  show    Shows group details\n```\n\n### `host` commands and options\n\nTo manage "groups" type `sshc host --help` to see options.  \n_example:_\n```console\n$ sshc host --help\nUsage: sshc host [OPTIONS] COMMAND [ARGS]...\n\n  Manage hosts\n\nOptions:\n  -h, --help  Show this message and exit.\n\nCommands:\n  create  Create new host configuration\n  delete  Delete host from configuration\n  list    List hosts\n  set     Changes/sets host configuration parameters\n  show    Display info for host\n  test    Test SSH host connection\n```\n\n### Output styling and user ENV variables\n\n`sshc host show` can display host output is several formats, you can specify it with `sshc host show <host> --style <style>`  \nAvailable styles are:\n\n| Style              | Description                                       |\n|--------------------|---------------------------------------------------|\n| `panels` (default) | Display data in several panels                    |\n| `card`             | Add data to single "card"                         |\n| `simple`           | Simple output with minimal decorations            |\n| `table`            | Flat table with 3 columns                         |\n| `table2`           | Nested table with separated host SSH params       |\n| `json`             | JSON output, useful for binding with other tools |\n\nIf you want to have some style statically set for your shell, you can export ENV variable with `export SSHC_HOST_STYLE=table`, and add it to `.profile` or `.bashrc` or `.zshrc`, so its set when shell session is starting, to set "default" style to that one.\n\nIn case user do not line "fancy" colors in output, you can set ENV variable to disable all color outputs with `export NO_COLOR=1`. If you want it permanently you can add it to startup "rc" files as well.\n\n\n> NOTE! When sending output into non-terminal such as to file, SSHClick will recognize that and will remove all ANSI Escape characters (colors and stuff...) so that output is captured in clear way.\n\n\n## Recorded demos\n\nFollowing demos will use this config sample file as input (located in ~/.ssh/config):\n\n```\n#<<<<< SSH Config file managed by sshclick >>>>>\n\n#-------------------------------------------------------------------------------\n#@group: network\n#@desc: Network devices in my lab\n#@info: user=\'admin\' password=\'password\'\n#@info: Not really, but for demo its ok :)\n#-------------------------------------------------------------------------------\nHost net-switch1\n    hostname 10.1.1.1\n\nHost net-switch2\n    hostname 10.1.1.2\n\nHost net-switch3\n    hostname 10.1.1.3\n\nHost net-*\n    user admin\n\n\n#-------------------------------------------------------------------------------\n#@group: jumphost\n#@desc: Edge-server / SSH bastion\n#@info: Used for jump-proxy from intnet to internal lab servers\n#-------------------------------------------------------------------------------\nHost jumper1\n    hostname 123.123.123.123\n    user master\n    port 1234\n\n\n#-------------------------------------------------------------------------------\n#@group: lab-servers\n#@desc: Testing/Support servers\n#@info: Some [red]important[/] detail here!\n#@info: We can have color markups in descriptions and info lines\n#-------------------------------------------------------------------------------\nHost lab-serv1\n    hostname 10.10.0.1\n    user admin\n\nHost lab-serv2\n    hostname 10.16.141\n\nHost lab-*\n    user user123\n    proxyjump jumper1\n\nHost server-behind-lab\n    hostname 10.30.0.1\n    user testuser\n    port 1234\n    proxyjump lab-serv1\n\n```\n\n### Demo showing some `sshc group` operations:\n[![asciicast](https://asciinema.org/a/BQoVXv2HSeIvTyATeKUBGfr89.svg)](https://asciinema.org/a/BQoVXv2HSeIvTyATeKUBGfr89)\n\n\n### Demo showing some `sshc host` operations:\n[![asciicast](https://asciinema.org/a/wzLefl49CRErBoFwC6ir96FFA.svg)](https://asciinema.org/a/wzLefl49CRErBoFwC6ir96FFA)\n\n\n### Demo displaying "end-to-end" tunnel visualization in graph\n[![asciicast](https://asciinema.org/a/I4O2bfDiRAN7xEGdTB1S9rChE.svg)](https://asciinema.org/a/I4O2bfDiRAN7xEGdTB1S9rChE)\n\n\n## Author\n\nKarlo Tisaj  \nemail: karlot@gmail.com  \ngithub: https://github.com/karlot\n\n\n## License\n\n[MIT License](LICENSE)\n\n\n',
    'author': 'Karlo Tisaj',
    'author_email': 'karlot@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/karlot/sshclick',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re
import fnmatch

from ..globals import *
from .ssh_host import SSH_Host
from .ssh_group import SSH_Group

import logging
logging.basicConfig(level=logging.INFO)

class SSH_Config:
    """
    SSH Configuration class

    Main class for handling SSH configuration, reading from file, parsing and
    generating and writing contents back to SSH configuration file
    """

    def __init__(self, file: str, config_lines: list = [], stdout: bool = DEFAULT_STDOUT):
        self.ssh_config_file: str = file
        self.ssh_config_lines: list = config_lines

        # configuration representation (array of SSH groups?)
        self.groups: list = [SSH_Group(name=DEFAULT_GROUP_NAME, desc="Default group")]

        # options
        self.stdout: bool = stdout

        # parsing "cache" info
        self.current_grindex: int = 0
        self.current_group: str = DEFAULT_GROUP_NAME
        self.current_host: SSH_Host = None
        self.current_host_info: list = []


    def read(self):
        """
        Read content of SSH config file
        """
        with open(self.ssh_config_file) as fh:
            self.ssh_config_lines = fh.readlines()
        return self


    def _config_flush_host(self) -> None:
        """
        Internal function used to flush host configuration while parsing config file
        """
        if self.current_host:
            if self.current_host.type == "normal":
                self.groups[self.current_grindex].hosts.append(self.current_host)
            else:
                self.groups[self.current_grindex].patterns.append(self.current_host)
            # Reset "cache" since we flushed host info
            self.current_host = None


    def parse(self):
        """
        Parse config lines one by one and generate configuration structure
        """
        # Parse each line of the configuration, line by line
        for config_line_index, line in enumerate(self.ssh_config_lines):
            line = line.strip()     # remove start and end whitespace

            # Skip empty lines, go to next...
            if not line:
                continue
            
            # Process special metadata for grouping: "#@group", "#@desc", "#@info"
            if line.startswith("#@"):
                match = re.search(r"^#@(\w+):\s*(.+)$", line)

                # In case special comment is unreadable
                if not match:
                    logging.warning(f"META: Unmatched metadata line {line} on SSH-config line number {config_line_index}")
                    continue
                
                metadata, value = match.groups()
                if metadata == "group":
                    # New group found... flush any previous data and create new baseline
                    self._config_flush_host()

                    logging.debug(f"META: Starting new group: {value}")
                    self.groups.append(SSH_Group(name=value))

                    self.current_grindex = len(self.groups) - 1
                    self.current_group = value
                    logging.debug(f"META: Group index set to: {self.current_grindex}")
                    continue

                elif metadata == "desc":
                    logging.debug(f"META: Setting 'desc' param to '{value}' for group '{self.groups[self.current_grindex].name}'")
                    self.groups[self.current_grindex].desc = value
                    continue

                elif metadata == "info":
                    logging.debug(f"META: Setting 'info' param to '{value}' for group '{self.groups[self.current_grindex].name}'")
                    self.groups[self.current_grindex].info.append(value)
                    continue

                elif metadata == "host":
                    logging.debug(f"META: Host comment found '{value}' Caching for next host definition...'")
                    self.current_host_info.append(value)
                    continue

                else:
                    logging.warning(f"META: Unhandled metadata metadata '{metadata}' on SSH-config line number {config_line_index}")
                    continue

            # Ignore rest of commented lines
            if line.startswith("#"):
                logging.debug(f"DROP COMMENT: '{line}'")
                continue

            # Here we expect only normal ssh config lines "Host" is usually the keyword that begins definition
            # if we find any other keyword before first host keyword is defined, configuration is wrong probably
            match = re.search(r"^(\w+)\s+(.+)$", line)
            if not match:
                logging.warning(f"KEYWORD: Incorrect configuration line '{line}' on SSH-config line number {config_line_index}")
                continue

            keyword, value = match.groups()
            keyword = keyword.lower()         # keywords are case insensitive, so we lowercase them

            # --- Found "host" keyword, that defines new block, usually followed with name
            if keyword == "host":
                self._config_flush_host()

                host_type = "pattern" if "*" in value else "normal"
                logging.debug(f"Host '{value}' is '{host_type}' type!")

                self.current_host = SSH_Host(name=value, group=self.current_group, type=host_type, info=self.current_host_info)

                # Reset global host info cache when we find new host (from this line, any host comments will apply to next host)
                self.current_host_info = []
                continue
            else:
                # any other normal line we just use as it is, wrong or not... :)
                # Currently there is no support for keyword validation
                if not self.current_host:
                    logging.warning(f"Config info without Host definition on SSH-config line number {config_line_index}")
                    exit(1)
                else:
                    logging.debug(f"Config keyword for host '{self.current_host}': {keyword} -> {value}")
                    # Save any specific info...
                    if keyword in PARAMS_WITH_ALLOWED_MULTIPLE_VALUES:
                        if not keyword in self.current_host.params:
                            self.current_host.params[keyword] = [value]
                        else:
                            self.current_host.params[keyword].append(value)
                    else:
                        self.current_host.params[keyword] = value
                        continue
        
        # Last entries must be flushed manually as there are no new "hosts" to trigger storing parsed data into config struct
        self._config_flush_host()
        return self


    def generate_ssh_config(self):
        """
        SSH config generation function

        Function takes config data structure, and target file information.
        Then generates SSH config compatible file with all data, compatible with
        internal object model.
        """
        SSHCONFIG_INDENT = "    "

        # First we lines before we flush them into file
        lines = ["#<<<<< SSH Config file managed by sshclick >>>>>\n"]

        for group in self.groups:
            # Ship default group as it does not have to be specified
            render_header = False if group.name == DEFAULT_GROUP_NAME else True
            
            if render_header:
                # Add extra blank line when outputting new group header
                lines.append("\n")
                # Start header line for the group with known metadata
                lines.append(f"#{'-'*79}\n")        # add horizontal decoration line
                lines.append(f"#@group: {group.name}\n")
                if group.desc:
                    lines.append(f"#@desc: {group.desc}\n")
                for info in group.info:
                    lines.append(f"#@info: {info}\n")
                lines.append(f"#{'-'*79}\n")        # add horizontal decoration line

            # Append hosts and patterns items from group
            for host in group.hosts + group.patterns:
                # If there is host-info assigned to host, add it before adding "host" definition
                for host_info in host.info:
                    lines.append(f"#@host: {host_info}\n")
                
                # Add "host" line definition
                lines.append(f"Host {host.name}\n")

                # Add all assigned host params
                for token, value in host.params.items():
                    if type(value) is str:
                        lines.append(f"{SSHCONFIG_INDENT}{token} {value}\n")
                    elif type(value) is list:
                        for v in value:
                            lines.append(f"{SSHCONFIG_INDENT}{token} {v}\n")
                    else:
                        raise Exception("Host parameter is not 'str' or 'list'!!!")
                
                # Add newline after host definition
                lines.append("\n")
            
        # Store output lines
        self.ssh_config_lines = lines
        return self
        

    def write_out(self):
        """
        Write generated SSH config to target file
        """
        if self.stdout:
            print("".join(self.ssh_config_lines))
        else:
            with open(self.ssh_config_file, "w") as out:
                out.writelines(self.ssh_config_lines)
        

    def find_group_by_name(self, name: str, throw_on_fail: bool = True):
        """
        Find group in configuration that matches the name (strict match, one only!)
        On success returns matched group, on fail depending on 'throw_on_fail' flag
        function will either return 'None' or will throw exception
        """
        for group in self.groups:
            if group.name == name:
                return group
        if not throw_on_fail:
            return None
        raise Exception(f"Requested group '{name}' not found in the SSH configuration")


    def find_host_by_name(self, name: str, throw_on_fail: bool = True):
        """
        Find host in configuration that matches the name (strict match, one only!)
        On success returns host and his assigned group, on fail depending on 'throw_on_fail' flag
        function will either return ('None','None') or will throw exception
        """
        found_host = None
        found_group = None
        
        for group in self.groups:
            all_hosts = group.hosts + group.patterns
            for host in all_hosts:
                if host.name == name:
                    found_host = host
                    found_group = group
        
        if not found_host and throw_on_fail:
            raise Exception(f"Requested host '{name}' not found in the SSH configuration")

        return found_host, found_group


    def get_all_host_names(self) -> list:
        """
        Return all host names from current configuration
        Useful for auto-completion, or for quick checking if name already exists
        """
        all_hosts: list = []
        for group in self.groups:
            for host in group.hosts:
                all_hosts.append(host.name)
        return all_hosts


    def get_all_group_names(self) -> list:
        """
        Return all group names from current configuration
        Useful for auto-completion, or for quick checking if name already exists
        """
        all_groups: list = []
        for group in self.groups:
            all_groups.append(group.name)
        return all_groups


    def find_inherited_params(self, host_name: str) -> list:
        """
        Returns array of 2-item tuples, where first item is name of pattern from
        which params are inherited, and second item is parameters dictionary from the pattern
        """
        inherited: list = []
        for group in self.groups:
            for pattern in group.patterns:
                # Check if any one of pattern (from all groups) will match host name
                if fnmatch.fnmatch(host_name, pattern.name):
                    inherited.append((pattern.name, pattern.params))

        return inherited




import imp
import click
from sshclick.globals import *
from sshclick.sshc import SSH_Config, SSH_Group, complete_ssh_host_names


#TODO: Check click.edit for multiline edit option (info, or even params?)

#------------------------------------------------------------------------------
# COMMAND: host set
#------------------------------------------------------------------------------
@click.command(name="set")
@click.option("-g", "--group", "target_group_name", default=None, help="Changes group for host")
@click.option("-r", "--rename", default=None, help="Rename host")
@click.option("-p", "--parameter", default=[], multiple=True, help="Sets parameter for the host, must be in 'param=value' format, to unset/remove parameter from host, set it to empty value (example: param='')")
@click.option("-i", "--info", default=None, multiple=True, help="Set host info, can be set multiple times, or set to empty value to clear it (example: info='')")
@click.option("--force", is_flag=True, default=False, help="Forces moving host to group that does not exist, by creating new group, and moving host to that group.")
@click.argument("names", nargs=-1, shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx, names, target_group_name, rename, parameter, info, force):
    """
    Changes/sets host(s) configuration parameters

    Multiple hosts can be specified, and can be set (moved) to different group, set/unset
    parameter or information lines. It is only not possible to "rename" multiple hosts!
    """

    config: SSH_Config = ctx.obj

    # Nothing was provided
    if not target_group_name and not rename and not parameter and not info:
        print("Calling set without setting anything is not valid on host(s). Run with 'sshc host set -h' for help.")
        ctx.exit(1)

    # When multiple hosts are defined, we cannot rename, as it makes no sense 
    if len(names) > 1:
        if rename:
            print(f"Cannot rename multiple hosts ({len(names)}) to same name ({rename}) is not supported")
            ctx.exit(1)

    for name in names:
        found_host, found_group = config.find_host_by_name(name, throw_on_fail=False)
        if not found_host:
            print(f"Cannot set anything on host '{name}' as it is not defined in configuration!")
            ctx.exit(1)
        
        # Move host to different group
        if target_group_name:
            # Find target group
            target_group = config.find_group_by_name(target_group_name, throw_on_fail=False)
            if not target_group:
                if not force:
                    print(f"Cannot move host '{name}' to group '{target_group_name}' which does not exist!")
                    print("Consider using --force to automatically create target group.")
                    exit(1)
                else:
                    # Create new group
                    new_group = SSH_Group(name=target_group_name)
                    # Add new group to config and set it as target group
                    config.groups.append(new_group)
                    target_group = new_group

            # Add host to target group
            if found_host.type == "normal":
                target_group.hosts.append(found_host)
                found_group.hosts.remove(found_host)
            else:
                target_group.patterns.append(found_host)
                found_group.patterns.remove(found_host)
            
        # Rename a host
        if rename:
            found_host.name = rename

        # Add info
        if info:
            if len(info[0]) > 0:
                found_host.info = info
            else:
                found_host.info = []

        # Sets parameters for host (or erase them if value is provided as unset)
        for item in parameter:
            param, value = item.split("=")
            param = param.lower()            # lowercase keyword/param as they are case insensitive
            if value:
                if param in PARAMS_WITH_ALLOWED_MULTIPLE_VALUES:
                    if not param in found_host.params:
                        found_host.params[param] = [value]
                    else:
                        if value in found_host.params[param]:
                            print(f"Cannot add existing value '{value}' to host parameter '{param}' multiple times!")
                        else:
                            found_host.params[param].append(value)
                else:
                    found_host.params[param] = value
            else:
                if param in found_host.params:
                    found_host.params.pop(param)
                else:
                    print(f"Cannot unset parameter that is not defined!")

        if not config.stdout:
            print(f"Modified host: {name}")

    # Write out modified config
    config.generate_ssh_config().write_out()
